#!/usr/bin/perl

# a little bit HTMLテキスト編集ツール リトルビット / by オレンジ工房
# alb.cgi 2015-07-22版
# ※最新版、サポート：http://alb.orange-factory.com/
# ※このコメントは削除しないでください。
# ※改変は自己責任でお願いします。

use strict;
use utf8;
use CGI qw(-utf8);
use CGI::Session qw(ip-match);
our $CGISI = 'ALBSID';		# cookie name for session
our $CGISE = '30m';			# expired 30 minutes after
CGI::Session->name($CGISI);

use URI::URL;
use File::Basename;		# dirname, filename
use File::Spec;
use HTML::Parser;
use POSIX qw(strftime);
use Encode::Guess qw(euc-jp shiftjis 7bit-jis);

use lib qw(lib);
use JSON;

# --------------------------------------
our $cgi = CGI->new();

our $session = CGI::Session->new(undef, $cgi, {Directory=>'./tmp'});
$session->expire($CGISE);
our $config;	# 設定データ（設定ファイルより）
our %json;		# 結果JSON

MAIN: eval {

	# 設定ファイル読み込み
	$config = read_config();

	# 設定ファイルがなければ初期設定へ
	if (!$config) {
		$json{'mode'} = 'init';
	}

	# 設定ファイルがあれば
	else {
		my $mode = $cgi->param('mode') || '';

		$json{'url'} = $session->param('url') || $cgi->param('url');

		# ログイン中なら通常処理
		if ( $session->param('login') ) {
			$json{'cname'} = $config->{cname};

			# 編集開始、または保存のとき
			if ( $mode eq 'edit'|| $mode eq 'save') {
				eval {
					# 編集可能かチェック、saveだったら保存もする
					checkEdit($mode);
				};
				if ($@) {
					$json{'result'} = 0;
					$json{'message'} = 'エラーメッセージ: ' . $@;
				}
			}

			# 表示のみの時
			else {
				# 何もしない
				$json{'mode'} = '';
			}
		}

		# 未ログイン
		else {
			# 今ログインしてるところ
			if ( $cgi->param('mode') eq 'login' ) {
				my $pwd = $cgi->param('pwd');
				if (login_ok( $cgi->param('pwd'))) {
					$json{'mode'} = "";
				}
				else {
					$json{'message'} = "パスワードが違います。";
					$json{'mode'} = 'login';
				}
			}
			# それ以外はログインフォームへ
			else {
				$json{'mode'} = 'login';
			}
		}
	}
};
# エラーのとき
if ($@) {
	$json{'error'} = 1;
	$json{'message'} = $@;
}

# 最後はJSONデータを返す
print $session->header(
	-type			=> 'application/json',
	-charset		=> 'utf-8',
	'Cache-Control' => 'no-cache',
), qq|\n|, encode_json(\%json);

# --------------------------------------
# ログインチェック
sub login_ok {

	my $pwd = shift;
	my $result = 0;

	if ($pwd eq $config->{'pwd'}) {
		$result = 1;
		$session->param('login', 1);
	}
	else {
		$session->param('login', 0);
	}

	return $result;
}
# --------------------------------------
# 設定ファイル読み込み
sub read_config {

	open FH, "<:encoding(utf8)", 'config.cgi' or return 0;

	my %config;

	foreach (<FH>) {
		chomp();
		my($k,$v) = split(/\t/);
		$config{$k} = $v;
	}
	close FH;

	return \%config;
}
# --------------------------------------
# 編集可能か確認
# グローバル%jsonに結果とか投入
# evalトラップしているので、dieしてOK。
sub checkEdit {

	my $mode = shift;

	my $url = $cgi->param('url');			# 編集対象ページのURL
	# 末尾がスラッシュで終わっていたら、index.htmlだと思ってみる
	# TO DO この判断は雑。URLかどうか入力チェック
	if ($url =~ m{/$}) {
		$url .= "index.html";
	}
	my $cehckURL  = URI::URL->new($url);					# 編集対象
	my $scriptURL = URI::URL->new($ENV{'REQUEST_URI'});		# このスクリプト自身

	$json{'file'}	 = $cehckURL->path;
	$json{'csspath'} = File::Spec->abs2rel(dirname($scriptURL->path), dirname($cehckURL->path)) . '/alb_edit.css';

	# テキストファイル、かつ、拡張子がそれらしいかチェック
	my $file = File::Spec->abs2rel($cehckURL->path, dirname($scriptURL->path));
	if (!-T $file or $file !~ /s?html?$/) {
		$json{'message'} = 'ファイルを特定できませんでした。';
		return;
	}

	# 編集領域があるか
	# 文字コード推測
	$json{'encoding'} = guess($file);
	# ファイルを開いてパース
	my $encode = '<:encoding(' . $json{'encoding'} . ')';
	open my $fh, $encode, $file or die($!);
	&Editable::parse_file($config->{'cname'}, $fh);		# 探す
	close $fh;
	# 編集領域のリストが空
	if (@Editable::result == ()) {
		$json{'message'} = "編集が許可されていません。";
		return;
	}

	# ここまでたどり着けば編集可能
	$json{'mode'} = 'edit';
	$json{'result'} = 1;

	# 保存じゃないならここで終わり。
	return if ($mode ne 'save');

	# 保存
	my @replace = $cgi->param('edit');		# 受け取った書き換え内容

	# 編集領域の数を比較
	if (scalar(@Editable::result) != scalar(@replace)) {
		$json{'message'} = "編集箇所の数が一致しませんでした。";
		$json{'result'} = 0;
	}

	# HTMLファイル一気読み
	my $html;
	open $fh, $encode, $file or die($!);
	{
		local $/ = undef;
		$html = <$fh>;
	}
	close $fh;

	# 順に置き換え
	# ToDo (可能性は低いけど)タグも内容も含めて全く同じ編集領域ガ複数あったら誤動作する。
open DEBUG, ">debug_replace.txt";
print DEBUG $html, "\n", "-"x40, "\n\n";
	foreach my $i (0..$#replace) {
		# 送信されたデータの最低限のクリーンアップ
		$replace[$i] =~ s/\s+$//g;
		$replace[$i] =~ s/^\s+//g;
		$replace[$i] =~ s/<br>$//g;
		my $ed = $Editable::result[$i];
		my $search	= $ed->{'startHtml'} . $ed->{'innerHtml'} . $ed->{'endHtml'};	# 開始タグ-内容-閉じタグ
		my $replace = $ed->{'startHtml'} . $replace[$i] 	  . $ed->{'endHtml'};	# を、コレに置き換え
if ($html!~/$ed->{'innerHtml'}/) {
	print DEBUG "NOT MATCH\n", $ed->{'innerHtml'}, "\n\n";
}
print DEBUG $search, "\n", "-"x40, "\n\n", $replace, "\n", "-"x40, "\n\n";
		# 置き換え。失敗したらエラー。
		unless ($html =~ s/$search/$replace/) {
			die("編集箇所の置換に失敗しました。\n");
		}
	}
close DEBUG;

	# 上書き（と言いつつ上書きじゃないけど）

	# 書き換え後HTMLを一旦保存
	my $tmpfile = $file . '.T' . strftime("%Y%m%d%H%M%S", localtime());
	die('一時的に保存に失敗しました。もう一度、試してください。') if (-T $tmpfile);
	open $fh, ">:encoding($json{encoding})", $tmpfile or die("保存に失敗しました。このフォルダへの書き込みが許可されていないかもしれません。");
	print $fh $html;
	close $fh;

	# 元ファイルをよける
	my $bkupfile = $file . '.B' . strftime("%Y%m%d%H%M%S", localtime());
	rename $file, $bkupfile or die("保存に失敗しました。");

	# テンポラリファイルをリネーム（差し替え完了）
	rename $tmpfile, $file or die("保存に失敗しました。(!)ページが見れなくなっています。編集前データは別名で残っていますので、FTP等で確認してください。");

	# ここまできたら元ファイルを消す
	unlink $bkupfile;

	$json{'message'} = "上書き保存しました。";
	$json{'mode'} = 'save';
}

# --------------------------------------
# 文字コード推測
sub guess {

	my $file = shift;

	my $encode;

	open FH, '<', $file or die($!);
	my $dat;
	{
		local $/ = undef;
		$dat = <FH>;
	}
	close FH;

	$encode = guess_encoding($dat);
	ref($encode) or die("ファイルの文字コードが不明です。 $encode\n");

	return $encode->name;
}
# --------------------------------------
# 編集ブロックを探すパッケージ
{
	package Editable;

	our @tag_stack = ();
	our %res = ();
	our @result = ();	# 編集領域のリスト
	our $editable = 0;

	our $p;

	# --------------------------------------
	# パース
	sub parse_file {
		my $cname = shift;
		init($cname);
		$p->parse_file(@_);
	}

	# --------------------------------------
	# 初期化
	sub init {

		my $cname = shift;

		$p = HTML::Parser->new();

		# 開始タグを見つけたとき
		$p->handler('start', sub {

			my($tag, $text, $attr) = @_;

			# 回収中ならそのまま内容に含める
			if ($editable) {
				$res{'innerHtml'} .= $text;
				push @tag_stack, $tag;
			}
			# 回収中でなく、クラス名一致したら、回収開始
			elsif ($attr->{'class'} =~ /\b$cname\b/) {
				$editable++;
				push @tag_stack, $tag;
				$res{'startHtml'} = $text;
			}
		}, "tagname, text, attr");

		# 終了タグを見つけたとき
		$p->handler('end', sub{

			my($tag, $text) = @_;

			# 回収中なら（回収中じゃなければ無視）
			if ($editable) {
				# スタックから開始タグ取り出して、
				while (my $pre_tag = pop @tag_stack) {
					# 閉じタグと一致したら当て込み終了
					if ($pre_tag eq $tag) {
						$tag = "";
						last;
					}
				}
				# スタックが残っていれば内容に含める
				if (@tag_stack != ()) {
					$res{'innerHtml'} .= $text;
				}
				# スタックが残ってない
				else {

					# 閉じタグが消えてない＝開始タグと対応取れてない
					die("編集領域を特定できませんでした。\n") if ($tag);

					# 対応取れてるなら、ここで回収終了。
					$editable = 0;
					$res{'endHtml'} = $text;
					$res{'innerHtml'} =~ s/([\¥\*\+\.\?\{\}\(\)\[\]\^\$\-\|])/\\$1/g;		# 正規表現マッチ用にエスケープ
					push @result, {
						startHtml => $res{'startHtml'},
						innerHtml => $res{'innerHtml'},
						endHtml   => $res{'endHtml'},
					};
					%res = ();
				}
			}
		}, "tagname, text");

		# ドキュメントが終わったとき
		# 回収フラグが立ったままだとおかしい。
		$p->handler('end_document', sub{ die('編集領域の設定が不正でした。') if ($editable) });

		# その他のものを見つけたとき
		# 単純に、回収中なら内容に含める
		$p->handler('text', sub{ $res{'innerHtml'} .= $_[0] if ($editable) }, "text");
		$p->handler('comment', sub{ $res{'innerHtml'} .= $_[0] if ($editable) }, "text");
		$p->handler('declaration', sub{ $res{'innerHtml'} .= $_[0] if ($editable) }, "text");
		$p->handler('process', sub{ $res{'innerHtml'} .= $_[0] if ($editable) }, "text");
	}
}
